
#ifndef __python_cnorxz_array_wrapper_h__
#define __python_cnorxz_array_wrapper_h__

#include <Python.h>
#include "cnorxz.h"
#include "cereal/cnorxz_cereal.h"
#include "numpy/ndarraytypes.h"

namespace CNORXZ
{

    class CArrayWrapperBase
    {
    public:
	DEFAULT_MEMBERS(CArrayWrapperBase);
	virtual ~CArrayWrapperBase() = default;

	virtual RangePtr range() const = 0;
	virtual SizeT size() const = 0;
	// virtual PyObject* get() const = 0; // operator[]!!
#ifdef HAVE_CEREAL
	virtual void writeFile(cer::Format f, const String& fname) const = 0;
	virtual void readFile(cer::Format f, const String& fname) const = 0;
#endif
	virtual int typenum() const = 0;
	virtual const void* data() const = 0;
	virtual SizeT datasize() const = 0;
    };

    class ArrayWrapperBase : public CArrayWrapperBase
    {
    public:
	virtual void* data() = 0;
    };

    template <typename T>
    struct Typenum
    {
	static const int value = NPY_NOTYPE;
    };

    template <>
    struct Typenum<Double>
    {
	static const int value = NPY_DOUBLE;
    };

    template <>
    struct Typenum<Int>
    {
	static const int value = NPY_INT32;
    };

    template <>
    struct Typenum<LInt>
    {
	static const int value = NPY_INT64;
    };

    template <>
    struct Typenum<SizeT>
    {
	static const int value = NPY_UINT64;
    };

    template <typename T>
    class ArrayWrapper : public ArrayWrapperBase
    {
    private:
	Sptr<ArrayBase<T>> mArr;
	    
    public:
	ArrayWrapper() : mArr( std::make_shared<MArray<T>>() ) {}
	ArrayWrapper(const RangePtr& r) : mArr( std::make_shared<MArray<T>>(r) ) {}
	ArrayWrapper(const Sptr<ArrayBase<T>>& arr) : mArr(arr) {}

	virtual RangePtr range() const override final { return mArr->range(); }
	virtual SizeT size() const override final { return mArr->size(); }
#ifdef HAVE_CEREAL
	virtual void writeFile(cer::Format f, const String& fname) const override final
	{
	    auto a = std::dynamic_pointer_cast<MArray<T>>( mArr );
	    if(a){
		if(f == cer::Format::XML) { cer::writeXMLFile(fname, *a); }
		else if(f == cer::Format::JSON) { cer::writeJSONFile(fname, *a); }
		else { cer::writeBINARYFile(fname, *a); }
	    }
	}
	
	virtual void readFile(cer::Format f, const String& fname) const override final
	{
	    auto a = std::dynamic_pointer_cast<MArray<T>>( mArr );
	    if(a){
		if(f == cer::Format::XML) { cer::readXMLFile(fname, *a); }
		else if(f == cer::Format::JSON) { cer::readJSONFile(fname, *a); }
		else { cer::readBINARYFile(fname, *a); }
	    }
	}
#endif
	virtual int typenum() const override final { return Typenum<T>::value; }
	virtual const void* data() const override final
	{ return reinterpret_cast<const void*>( mArr->data() ); }
	virtual void* data() override final
	{ return reinterpret_cast<void*>( mArr->data() ); }
	virtual SizeT datasize() const override final { return size() * sizeof(T); }
    };
	
}

struct PyCxArrayB
{
    PyObject_HEAD
    CNORXZ::CArrayWrapperBase* ptrObj;
};

int PyCxArrayB_init(PyCxArrayB* self, PyObject* args, PyObject* kwds);
void PyCxArrayB_dealloc(PyCxArrayB* self);
PyObject* PyCxArrayB_size(PyCxArrayB* self);
PyObject* PyCxArrayB_range(PyCxArrayB* self);
PyObject* PyCxArrayB_writeFile(PyCxArrayB* self, PyObject* args, PyObject* kwds);
PyObject* PyCxArrayB_readFile(PyCxArrayB* self, PyObject* args, PyObject* kwds);
PyTypeObject* PyCxArrayBType_init();

PyObject* PyCxReadFile(PyObject* self, PyObject* args, PyObject* kwds);

extern PyMethodDef PyCxArrayB_methods[];
extern PyTypeObject PyCxArrayBType;

#endif
