
#include "array_wrapper.h"
#include "range_wrapper.h"
#include "numpy/arrayobject.h"

using namespace CNORXZ;
using namespace CNORXZ::cer;

int PyCxArrayB_init(PyCxArrayB* self, PyObject* args, PyObject* kwds)
{
    static char* kwlist[] = { "type", "extension" , NULL };
    SizeT ext = 0;
    char* type = NULL;

    if(not PyArg_ParseTupleAndKeywords(args, kwds, "|sk", kwlist, &type, &ext)){
	return -1;
    }
	    
    if(ext != 0){
	RangePtr r = CRangeFactory(ext).create();
	// TODO: different types!!!
	self->ptrObj = new ArrayWrapper<Double>(r);
    }
    else {
	self->ptrObj = new ArrayWrapper<Double>();
    }
    return 0;
}

void PyCxArrayB_dealloc(PyCxArrayB* self)
{
    delete self->ptrObj;
    Py_TYPE(self)->tp_free(self);
}

PyObject* PyCxArrayB_size(PyCxArrayB* self)
{
    if(not self->ptrObj->range()){
	PyErr_SetString(PyExc_RuntimeError, "array not initialized");
	return NULL;
    }
    const SizeT retval = self->ptrObj->size();
    return Py_BuildValue("k", retval);
}

PyObject* PyCxArrayB_range(PyCxArrayB* self)
{
    if(not self->ptrObj->range()){
	PyErr_SetString(PyExc_RuntimeError, "array not initialized");
	return NULL;
    }
    const RangePtr r = self->ptrObj->range();
    PyCxRange* o;
    o = PyObject_New(PyCxRange, &PyCxRangeType);
    o->ptrObj = new RangeWrapper(r);
    return Py_BuildValue("O", o);
}

Format formatFromString(const char* fstr)
{
    Format f = Format::BINARY;
    if(fstr != NULL){
	if(strcmp(fstr,"binary") == 0){
	    f = Format::BINARY;
	}
	else if(strcmp(fstr,"json") == 0){
	    f = Format::JSON;
	}
	else if(strcmp(fstr,"xml") == 0){
	    f = Format::XML;
	}
	else {
	    PyErr_SetString(PyExc_RuntimeError, "unknown format");
	}
    }
    return f;
}

#ifdef HAVE_CEREAL

PyObject* PyCxArrayB_writeFile(PyCxArrayB* self, PyObject* args, PyObject* kwds)
{
    if(not self->ptrObj->range()){
	PyErr_SetString(PyExc_RuntimeError, "array not initialized");
    }
    static char* kwlist[] = { "format", "filename" , NULL };
    char* fname = NULL;
    char* format = NULL;

    if(not PyArg_ParseTupleAndKeywords(args, kwds, "ss|", kwlist, &fname, &format)){
	return NULL;
    }
    const String fn(fname);
    const Format f = formatFromString(format);
    self->ptrObj->writeFile(f, fn);
    Py_RETURN_NONE;
}

PyObject* PyCxArrayB_readFile(PyCxArrayB* self, PyObject* args, PyObject* kwds)
{
    static char* kwlist[] = { "format", "filename" , NULL };
    char* fname = NULL;
    char* format = NULL;

    if(not PyArg_ParseTupleAndKeywords(args, kwds, "ss|", kwlist, &fname, &format)){
	return NULL;
    }
    const String fn(fname);
    const Format f = formatFromString(format);
    self->ptrObj->readFile(f, fn);
    Py_RETURN_NONE;
}
#endif

PyObject* PyCxArrayB_npa(PyCxArrayB* self)
{
    // return viewing (!) numpy array
    if(not self->ptrObj->range()){
	PyErr_SetString(PyExc_RuntimeError, "array not initialized");
	return NULL;
    }
    const SizeT d = self->ptrObj->range()->dim();
    std::vector<npy_intp> dims(d);
    for(SizeT i = 0; i != dims.size(); ++i){
	dims[i] = self->ptrObj->range()->sub(i)->size();
    }
    // TODO: Do the following for mutable arrays!!!
    /*
    Py_INCREF(self); // keep data alive
    PyObject* o = PyArray_SimpleNewFromData
	( static_cast<int>( d ) , dims.data(), self->ptrObj->typenum(), self->ptrObj->data() );
    PyArray_BASE(o) = (PyObject*) self;
    */
    PyObject* o = PyArray_SimpleNew
	( static_cast<int>( d ) , dims.data(), self->ptrObj->typenum() );
    void* data = PyArray_DATA(o);
    //const npy_intp s = PyArray_NBYTES(o);
    const SizeT s = static_cast<SizeT>( PyArray_NBYTES(o) );
    if(s != self->ptrObj->datasize()) {
	PyErr_Format(PyExc_RuntimeError, "lib error %d vs %d", s, self->ptrObj->datasize());
	//PyErr_SetString(PyExc_RuntimeError, "lib error");
	return NULL;
    }
    memcpy( data, self->ptrObj->data(), self->ptrObj->datasize() );
    return o;
}

PyMethodDef PyCxArrayB_methods[] = {
    { "size", (PyCFunction) PyCxArrayB_size, METH_VARARGS| METH_KEYWORDS, "return size of the array" },
    { "range", (PyCFunction) PyCxArrayB_range, METH_VARARGS| METH_KEYWORDS, "return array range" },
    { "writeFile", (PyCFunction) PyCxArrayB_writeFile, METH_VARARGS| METH_KEYWORDS, "write array to file" },
    { "readFile", (PyCFunction) PyCxArrayB_readFile, METH_VARARGS| METH_KEYWORDS, "read array from file" },
    { "npa", (PyCFunction) PyCxArrayB_npa, METH_VARARGS| METH_KEYWORDS, "return numpy array (view)" },
    { NULL, NULL, 0, NULL }
};

PyTypeObject PyCxArrayBType = { PyVarObject_HEAD_INIT(NULL,0) "cnorxz.Array" };

PyTypeObject* PyCxArrayBType_init()
{
    PyCxArrayBType.tp_new = PyType_GenericNew;
    PyCxArrayBType.tp_basicsize = sizeof(PyCxArrayB);
    PyCxArrayBType.tp_dealloc = (destructor) PyCxArrayB_dealloc;
    PyCxArrayBType.tp_flags = Py_TPFLAGS_DEFAULT;
    PyCxArrayBType.tp_doc = "cnorxz Array wrapper";
    PyCxArrayBType.tp_methods = PyCxArrayB_methods;
    PyCxArrayBType.tp_init = (initproc) PyCxArrayB_init;
    
    import_array();
    
    if(PyType_Ready(&PyCxArrayBType) < 0) {
	return NULL;
    }
    else {
	return &PyCxArrayBType;
    }
}

PyObject* PyCxReadFile(PyObject* self, PyObject* args, PyObject* kwds)
{
    PyCxArrayB* out = (PyCxArrayB*) PyObject_NEW(PyCxArrayB, &PyCxArrayBType);
    out->ptrObj = new ArrayWrapper<Double>();
    PyCxArrayB_readFile(out,args,kwds);
    return Py_BuildValue("O", out);;
}
